#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include "x11_simple.h"
#include BUFFER

#define MAINLOOP_POLL_DELAY 10

#include "display/common.c"

static Display *dsp = NULL;
static XColor xcolors[16];
static Window wnd;
static GC gc;
static XFontStruct *fnt;
static int charw, charh;
static int windw, windh;
static int ascent, descent;
static int page;
static Atom wmDeleteWindow;
static Atom wmNormalHints;
static Colormap colormap;

/* begin and end of selection area
 * offset from top left visible cell */
static int sel_begin = -1, sel_end = -1;

/* mouse pointer x;y coordinates where selection began */
static int sel_begin_x = -1;
static int sel_begin_y = -1;


static int sel_mode = 0; /* if 1, selection mode (w/ mouse drag) is active */

static unsigned char *sel_buf = NULL; /* selection buffer */
static unsigned int sel_buflen = 0;   /* length of selection buffer */

watched watches[MAXWATCHED];
int numwatched = 0;

void *dsp_bf;

/**
 * set active buffer
 */
void display_set_buffer(void *b)
{
	dsp_bf = b;
	display_update(0, 0, windw - 1, windh - 1);
}

static inline void display_cell(const int xc, const int yc, buf_cell_t c)
{
	unsigned char tmp;
	if (c.c == 0) {
		XSetForeground(dsp, gc, xcolors[c.color.bg].pixel);
		XFillRectangle(dsp, wnd, gc, xc, yc, charw, charh);
	} else {
		if (c.color.attr.reverse) {
			tmp = c.color.bg;
			c.color.bg = c.color.fg;
			c.color.fg = tmp;
		}

		if (c.color.attr.underline)
			c.color.fg = 15; // bufstate.underline_fg;
		else if (c.color.attr.halfbright)
			c.color.fg = 7; // bufstate.halfbright_fg;
		else if (c.color.attr.bold)
			c.color.fg |= 0x08;
		if (c.color.attr.blink)
			c.color.bg |= 0x08; // highlight instead of blink
		XSetForeground(dsp, gc, xcolors[c.color.bg].pixel);
		XFillRectangle(dsp, wnd, gc, xc, yc, charw, charh);
		XSetForeground(dsp, gc, xcolors[c.color.fg].pixel);
		XDrawString(dsp, wnd, gc, xc, yc + ascent, &c.c, 1);
	}
}


/**
 * redraw part of display
 */
void display_update(const int x1, const int y1, const int x2, const int y2)
{
	int x, y;
	int yc, xc;
	buf_cell_t c;
	buf_state_t *bfs;

	int off;

	bfs = buf_get_state(dsp_bf);

	XSetFont(dsp, gc, fnt->fid);

	yc = y1 * charh;
	for(y = y1; y <= y2; y++) {
		buf_set_r(dsp_bf, x1, y);
		xc = x1 * charw;

		off = y * windw + x1;

		for(x = x1; x <= x2; x++) {
			c = buf_read_cell(dsp_bf);
			display_cell(xc, yc, c);

			/* reverse cursor area, when selection mode is off
			 * also when we're inside the selection area */
			if((!sel_mode && bfs->cursor_type == buf_cursor_block && x == bfs->cursor_x && y-bfs->scrv == bfs->cursor_y) ||
					(off >= sel_begin && off <= sel_end) ||
					(off >= sel_end && off <= sel_begin)) {
				XSetFunction(dsp, gc, GXxor);
				XSetForeground(dsp, gc, xcolors[7].pixel);
				XFillRectangle(dsp, wnd, gc, xc, yc, charw, charh);
				XSetFunction(dsp, gc, GXcopy);
			}
			xc += charw;
			off ++;
		}

		yc += charh;
	}
}

/**
 * shift an area of the display
 */
void display_shift(const int x1, const int y1, const int x2, const int y2, const int xd, const int yd)
{
	XCopyArea(dsp, wnd, wnd, gc, x1 * charw, y1 * charh, (x2-x1+1) * charw, (y2-y1+1) * charh, (x1+xd) * charw, (y1+yd) * charh);
}

/**
 * scroll display one line up
 */
void display_scroll()
{
	if(sel_begin != -1) {
		sel_begin -= windw;
		sel_end -= windw;
		if(sel_begin < 0 || sel_end < 0) {
			sel_begin = -1;
			sel_end = -1;
			/* TODO: update only as necessary */
			display_update(0, 0, windw-1, windh-1);
			return;
		}
	}

	XCopyArea(dsp, wnd, wnd, gc, 0, charh, windw * charw, (windh-1) * charh, 0, 0);

	display_update(0, windh-1, windw-1, windh-1);
}

/**
 * hide cursor
 */
void display_cursor_hide()
{
	buf_state_t *bfs;
	bfs = buf_get_state(dsp_bf);
	bfs->cursor_x --;
	display_update(bfs->cursor_x + 1, bfs->cursor_y, bfs->cursor_x + 1, bfs->cursor_y);
	bfs->cursor_x ++;
}

static char *display_title = NULL;

/**
 * set display title
 */
void display_set_title(char *title)
{
	if (display_title != NULL)
		free(display_title);
	display_title = strdup(title);
	if (dsp != NULL)
		XStoreName(dsp, wnd, title);
}

static void *bell_timer = NULL;

static void display_bell_callback(display_timer_cmd_t cmd, void *cookie)
{
	char *s;

	switch(cmd) {
		case timer_created:
			if (display_title == NULL)
				s = strdup(CFG_DISPLAY_BELL_TEXT);
			else {
				s = malloc(strlen(display_title) + strlen(CFG_DISPLAY_BELL_TEXT) + 2);
				sprintf(s, "%s %s", display_title, CFG_DISPLAY_BELL_TEXT);
			}
			XStoreName(dsp, wnd, s);
			free(s);
			break;
		case timer_run:
			XStoreName(dsp, wnd, display_title == NULL ? "" : display_title);
			break;
		case timer_free:
			bell_timer = NULL;
			break;
	}
}

void display_bell()
{
	if (bell_timer == NULL)
		bell_timer = display_timer_create(CFG_DISPLAY_BELL_LENGTH, display_bell_callback, NULL);
	else
		display_timer_set_ttl(bell_timer, CFG_DISPLAY_BELL_LENGTH);
}

/* Count mouse clicks */
void *click_timer;
int click_count;
static void display_reset_click_callback(display_timer_cmd_t cmd, void *cookie)
{
	switch(cmd) {
		case timer_run:
			log_printf(3, "resetting click count.\n");
			click_count = 0;
			break;
		case timer_free:
			click_timer = NULL;
			break;
		case timer_created:
			/* Do nothing */
			;
	}
}

/**
 * Process keyrelease
 */
static void keyrelease(XKeyEvent *xk)
{
	int meta, r;
	char kbuf[32];
	KeySym keysym;

	meta = xk->state & Mod1Mask;

	r = XLookupString(xk, kbuf, 32, &keysym, NULL);
	if (keysym == XK_Alt_L)
		display_numcode(-1);
}

/* respond to XSelectionNotify */
static void selection_paste(Window win, unsigned int prop, int Delete)
{
    long            nread;
    unsigned long   bytes_after;
    XTextProperty   ct;

    if (prop == None) return;
    for (nread = 0, bytes_after = 1; bytes_after > 0; nread += ct.nitems) {
		if ((XGetWindowProperty(dsp, win, prop, (nread / 4), 4096,
					Delete, AnyPropertyType, &ct.encoding,
					&ct.format, &ct.nitems, &bytes_after,
					&ct.value) != Success)) {
			XFree(ct.value);
			return;
		}
		if (ct.value == NULL) continue;

		int i;
		for(i=0; i<ct.nitems; i++)
			buf_dsp_char(dsp_bf, ct.value[i]);

		XFree(ct.value);
    }
}

/* Paste from the 'clipboard' */
void dsp_paste_selection()
{
	Window Sown = XGetSelectionOwner (dsp, XA_PRIMARY);
	int i;
	if(Sown == wnd) {
		/* we own the selection */
		for(i=0; i<sel_buflen; i++) buf_dsp_char(dsp_bf, sel_buf[i]);
		return;
	}

	if (Sown != None) {
		/* send request for selection */
		Atom prop = XInternAtom(dsp, "VT_SELECTION", False);
		XConvertSelection (dsp, XA_PRIMARY, XA_STRING, prop, wnd, CurrentTime);
		XFlush (dsp);
	} else {
		selection_paste(DefaultRootWindow(dsp), XA_CUT_BUFFER0, False);
	}
}


/* Copy to the clipboard */
void dsp_copy_selection()
{
	if(sel_begin == -1) {
		sel_buflen = 0;
		XSetSelectionOwner(dsp, XA_PRIMARY, wnd, CurrentTime);
		return;
	}
	if(sel_end < sel_begin) {
		int s = sel_end;
		sel_end = sel_begin;
		sel_begin = s;
	}

	if(sel_buf) free(sel_buf);
	sel_buf = (unsigned char *) malloc(sel_end - sel_begin + 1);
	if(!sel_buf) abort();
	unsigned char *p = sel_buf;

	int y = sel_begin / windw;
	int x = sel_begin % windw;
	int i;

	buf_set_r(dsp_bf, x, y);

	for(i=sel_begin; i<=sel_end; i++) {
		buf_cell_t ch = buf_read_cell(dsp_bf);

		if(ch.c == 0) {
			x = 0; y++;
			buf_set_r(dsp_bf, x, y);
			*p++ = '\n';
			i = y*windw+x;
			continue;
		}

		*p++ = ch.c;

		x++;
		if(x == windw) {
			x = 0;
			y ++;
			buf_set_r(dsp_bf, x, y);
		}
	}

	sel_buflen = p - sel_buf;

	XSetSelectionOwner(dsp, XA_PRIMARY, wnd, CurrentTime);
}

/**
 * Process keypress
 */
static void keypress(XKeyEvent *xk)
{
	int r, i;
	int shft, ctrl, meta;
	char kbuf[32];
	KeySym keysym;

	shft = xk->state & ShiftMask;
	ctrl = xk->state & ControlMask;
	meta = xk->state & Mod1Mask;

	if(meta) {
		r = XLookupString(xk, kbuf, 32, &keysym, NULL);
		switch(keysym) {
			case XK_KP_Up:    display_numcode(8); return;
			case XK_KP_Down:  display_numcode(2); return;
			case XK_KP_Right: display_numcode(6); return;
			case XK_KP_Left:  display_numcode(4); return;
			case XK_KP_Prior: display_numcode(9); return;
			case XK_KP_Next:  display_numcode(3); return;
			case XK_KP_Home:  display_numcode(7); return;
			case XK_KP_End:   display_numcode(1); return;
			case XK_KP_Begin: display_numcode(5); return;
			case XK_KP_Insert: display_numcode(0); return;

			case XK_Up:
				buf_select(dsp_bf = buf_create());
				return;
			case XK_Left:
				dsp_bf = buf_switch_rel(-1);
				return;
			case XK_Right:
				dsp_bf = buf_switch_rel(1);
				return;
		}
		buf_dsp_char(dsp_bf, '\033');
		for(i=0;i<r;i++)
			buf_dsp_char(dsp_bf, kbuf[i]);
		return;
	}

	r = XLookupString(xk, kbuf, 32, &keysym, NULL);

	switch(keysym) {
	case XK_Tab:
		if(shft) {
			buf_dsp_send(dsp_bf, "\033[Z");
			return;
		}
		break;

	case XK_KP_Up:
	case XK_KP_Down:
	case XK_KP_Right:
	case XK_KP_Left:
		keysym = XK_Left + (keysym - XK_KP_Left);

	case XK_Up:
	case XK_Down:
	case XK_Right:
	case XK_Left:
		buf_dsp_char(dsp_bf, '\033');
		if(shft) {
			buf_dsp_char(dsp_bf, '[');
			buf_dsp_char(dsp_bf, "dacb"[keysym - XK_Left]);
			return;
		} else if(ctrl) {
			buf_dsp_char(dsp_bf, 'O');
			buf_dsp_char(dsp_bf, "dacb"[keysym - XK_Left]);
		} else {
			buf_dsp_char(dsp_bf, '[');
			buf_dsp_char(dsp_bf, "DACB"[keysym - XK_Left]);
		}
		return;

	case XK_KP_Prior:
	case XK_Prior:
		if(shft) {
			buf_adjust_scroll(dsp_bf, page);
			return;
		}

		buf_dsp_send(dsp_bf, "\033[5~");
		return;

	case XK_KP_Next:
	case XK_Next:
		if(shft) {
			buf_adjust_scroll(dsp_bf, -page);
			return;
		}
		buf_dsp_send(dsp_bf, "\033[6~");
		return;

	case XK_KP_Home:
	case XK_Home:
		buf_dsp_send(dsp_bf, "\033[1~");
		return;

	case XK_KP_End:
	case XK_End:
		buf_dsp_send(dsp_bf, "\033[4~");
		return;

	case XK_KP_Insert:
		if (shft) {
			buf_dsp_send(dsp_bf, "\033Op");
			return;
		}
	case XK_Insert:
		if (shft)
			dsp_paste_selection();
		else
			buf_dsp_send(dsp_bf, "\033[2~");
		return;

	case XK_KP_Delete:
		if (shft) {
			buf_dsp_send(dsp_bf, "\033On");
			return;
		}
	case XK_Delete:
		buf_dsp_send(dsp_bf, "\033[3~");
		return;

	case XK_F1:
	case XK_F2:
	case XK_F3:
	case XK_F4:
	case XK_F5:
		buf_dsp_send(dsp_bf, "\033[[");
		buf_dsp_char(dsp_bf, 'A' + (keysym - XK_F1));
		return;

	case XK_F6:
	case XK_F7:
	case XK_F8:
		buf_dsp_send(dsp_bf, "\033[1");
		buf_dsp_char(dsp_bf, '7' + (keysym - XK_F6));
		buf_dsp_char(dsp_bf, '~');
		return;

	case XK_F9:
	case XK_F10:
		buf_dsp_send(dsp_bf, "\033[2");
		buf_dsp_char(dsp_bf, '0' + (keysym - XK_F9));
		buf_dsp_char(dsp_bf, '~');
		return;

	case XK_F11:
	case XK_F12:
	case XK_F13:
	case XK_F14:
		buf_dsp_send(dsp_bf, "\033[2");
		buf_dsp_char(dsp_bf, '3' + (keysym - XK_F11));
		buf_dsp_char(dsp_bf, '~');
		return;
	}


	if(r>0) {
		for(i=0;i<r;i++)
			buf_dsp_char(dsp_bf, kbuf[i]);
	}
}

void display_exit()
{
	if (dsp_bf != NULL)
		buf_dsp_ctrl(dsp_bf, ctrl_hup);
	exit(0);
}

void display_get_size(int *rows, int *cols)
{
	*rows = windh;
	*cols = windw;
}

/**
 * start display
 */
void display_start()
{
	int dir, i;
	XCharStruct xch;
	fd_set rfds;
	fd_set wfds;
	fd_set efds;
	struct timeval tv;
	int n;
	XEvent e;
	int x1,y1,x2,y2;
	int oldw, oldh;
	XSizeHints xsh;
	buf_state_t *bst;
	XColor xclr;
	int new_sel_end;

	dsp = XOpenDisplay(NULL);
	if(!dsp) {
		log_printf(1, "Can't open X display\n");
		abort();
	}

	colormap = DefaultColormap(dsp, 0);

	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_0, &xcolors[0], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_1, &xcolors[1], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_2, &xcolors[2], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_3, &xcolors[3], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_4, &xcolors[4], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_5, &xcolors[5], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_6, &xcolors[6], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_7, &xcolors[7], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_8, &xcolors[8], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_9, &xcolors[9], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_10, &xcolors[10], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_11, &xcolors[11], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_12, &xcolors[12], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_13, &xcolors[13], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_14, &xcolors[14], &xclr);
	XAllocNamedColor(dsp, colormap, CFG_DISPLAY_COLOR_15, &xcolors[15], &xclr);

#ifndef NDEBUG
	log_printf(3, "opening font: %s\n", CFG_DISPLAY_X11_FONT);
#endif
	fnt = XLoadQueryFont(dsp, CFG_DISPLAY_X11_FONT);
	if (fnt == NULL) {
		log_printf(1, "Error opening font: %s\n", CFG_DISPLAY_X11_FONT);
		abort();
	}


	XTextExtents(fnt, "W", 1, &dir, &ascent, &descent, &xch);
	assert(xch.width > 0);
	charw = xch.width;
	charh = ascent + descent;

	wnd = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp), 0, 0, charw * 80, charh * 25, 0, xcolors[0].pixel, xcolors[0].pixel);
	oldw = charw * 80;
	oldh = charh * 25;
	if (display_title != NULL)
		XStoreName(dsp, wnd, display_title);

	XSelectInput(dsp, wnd, KeyReleaseMask | KeyPressMask | ButtonPressMask | ButtonReleaseMask |
						   ButtonMotionMask | ExposureMask | FocusChangeMask |
						   StructureNotifyMask | SubstructureNotifyMask);
	XMapWindow(dsp, wnd);

	wmDeleteWindow = XInternAtom(dsp, "WM_DELETE_WINDOW", 0);
	XSetWMProtocols(dsp, wnd, &wmDeleteWindow, 1);

	wmNormalHints = XInternAtom(dsp, "WM_NORMAL_HINTS", 0);
	xsh.flags = PResizeInc;
	xsh.width_inc = charw;
	xsh.height_inc = charh;
	XSetWMNormalHints(dsp, wnd, &xsh);

	gc = XCreateGC(dsp, wnd, 0, NULL);

	XFlush(dsp);

	buf_set_screen_dim(80, 25);
	windw = 80;
	windh = 25;
	page = 25;
	dsp_bf = buf_create();
	buf_select(dsp_bf);

	buf_dsp_ctrl(dsp_bf, ctrl_resize);

	/* MAIN LOOP */
	for(;;) {
		/* process X events */
		while(XPending(dsp)) {
nextevent:
			XNextEvent(dsp, &e);

			bst = buf_get_state(dsp_bf);

			switch(e.type) {
			case ClientMessage:
				if(e.xclient.format == 32 && e.xclient.data.l[0] == wmDeleteWindow) {
					log_printf(4, "window close\n");
					exit(0);
				}
				break;
			case ButtonPress:
				if(e.xbutton.button == 2)
					dsp_paste_selection();
				else if(e.xbutton.button == 4)
					buf_adjust_scroll(dsp_bf, CFG_DISPLAY_WHEEL_SCROLL);
				else if(e.xbutton.button == 5)
					buf_adjust_scroll(dsp_bf, -CFG_DISPLAY_WHEEL_SCROLL);
				else if(e.xbutton.button == 1) {
					log_printf(4, "Click count: %d\n", click_count);
					switch(click_count) {
						case 0: /* single click so far */
							{
								/* start selection mode */
								/* get character cell */
								int y = e.xbutton.y / charh;
								int x = e.xbutton.x / charw;
								if(y < windh && x < windw) {
									sel_mode = 1;
									sel_begin = sel_end = y*windw+x;
									sel_begin_x = e.xbutton.x;
									sel_begin_y = e.xbutton.y;
									/* TODO: update only as necessary */
									display_update(0, 0, windw-1, windh-1);
								}
							}
						break;
						case 1: /* double click so far */
							log_printf(1, " -> double click!\n");
						break;
						case 2: /* triple click so far */
							log_printf(1, " -> triple click!\n");
						break;
					}
					click_count++;
					/* Set up a timer to count clicks */
					click_timer = display_timer_create(CFG_DISPLAY_CLICK_FREQ, display_reset_click_callback, NULL);
				}
				break;
			case ButtonRelease:
				if(e.xbutton.button == 1 && sel_mode) {
					if ((sel_begin_x == e.xbutton.x) && (sel_begin_y == e.xbutton.y)) {
						/* No motion, single click in place - abort selection */
						sel_begin = -1;
						sel_end   = -1;
					}
					else {
						click_count = 0;
						if (click_timer != NULL)
							display_timer_delete(click_timer);
					}

					/* end selection mode */
					dsp_copy_selection();
					sel_mode = 0;
					display_update(0, 0, windw-1, windh-1);
				}
				break;
			case MotionNotify:
				if(sel_mode) {
					int y = e.xbutton.y / charh;
					int x = e.xbutton.x / charw;
					if(y < windh && x < windw) {
						int r1 = sel_begin / windw;
						int r2 = sel_end / windw;
						new_sel_end = y*windw+x;
						if (sel_end != new_sel_end) {
							sel_end = new_sel_end;
							int r;
							if(r2 < r1) {
								r = r1;
								r1 = r2;
								r2 = r;
							}
							if(y < r1) r1 = y; else if(y > r2) r2 = y;
							display_update(0, r1, windw-1, r2);
						}
					}
				}
				break;
			case SelectionClear:
				/* TODO: update only as necessary */
				sel_mode = 0;
				sel_begin = -1;
				sel_end = -1;
				display_update(0, 0, windw-1, windh-1);
				break;
			case SelectionNotify:
				/* received selection data */
				selection_paste(e.xselection.requestor, e.xselection.property, True);
				break;
			case KeyPress:
				keypress(&e.xkey);
				break;
			case ConfigureNotify:
				if(e.xconfigure.width != oldw || e.xconfigure.height != oldh) {
					log_printf(2, "resize request: %d %d\n", e.xconfigure.width, e.xconfigure.height);
					oldw = e.xconfigure.width;
					oldh = e.xconfigure.height;
					windw = oldw / charw;
					page = windh = oldh / charh;
					buf_set_screen_dim(windw, windh);
					buf_dsp_ctrl(dsp_bf, ctrl_resize);
				}
				break;
			case SelectionRequest:
				if(sel_buf && sel_buflen > 0) {
					XSelectionRequestEvent *req = &(e.xselectionrequest);
					XEvent respond;
					if(req->target == XA_STRING) {
						XChangeProperty(dsp, req->requestor, req->property,
								XA_STRING, 8, PropModeReplace,
								sel_buf, sel_buflen);
						respond.xselection.property=req->property;
					} else {
						respond.xselection.property= None;
					}
					respond.xselection.type= SelectionNotify;
					respond.xselection.display= req->display;
					respond.xselection.requestor= req->requestor;
					respond.xselection.selection=req->selection;
					respond.xselection.target= req->target;
					respond.xselection.time = req->time;
					XSendEvent(dsp, req->requestor,0,0,&respond);
					XFlush(dsp);
				}
				break;
			case KeyRelease:
				keyrelease(&e.xkey);
				break;
			case Expose:
				x1 = e.xexpose.x / charw;
				y1 = e.xexpose.y / charh;
				x2 = x1 + (e.xexpose.width / charw) + 1;
				y2 = y1 + (e.xexpose.height / charh) + 1;
				if(x2 > bst->w-1) x2 = bst->w-1;
				if(y2 > bst->h-1) y2 = bst->h-1;
				display_update(x1, y1, x2, y2);
				break;
			case MapNotify:
				display_update(0, 0, bst->w-1, bst->h-1);
				break;
			}
		}

		/* i/o etc is run only when not in selection mode
		 * we also need to block until next X event */
		if(sel_mode) {
			while(!XPending(dsp)) {
				usleep(MAINLOOP_POLL_DELAY);
				display_timer_run();
			}
			goto nextevent;
		}

		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_ZERO(&efds);
		n = 0;
		tv.tv_sec = 0;
		tv.tv_usec = MAINLOOP_POLL_DELAY;

		for(i=0; i<numwatched; i++) {
			if(watches[i].wtype & 1) FD_SET(watches[i].fd, &rfds);
			if(watches[i].wtype & 2) FD_SET(watches[i].fd, &wfds);
			if(watches[i].wtype & 4) FD_SET(watches[i].fd, &efds);

			if(watches[i].fd > n) n = watches[i].fd;
		}

		if(select(n+1, &rfds, &wfds, &efds, &tv)) {
			for(i=0; i<numwatched; i++) {
				n = 0;
				if(FD_ISSET(watches[i].fd, &rfds)) n += 1;
				if(FD_ISSET(watches[i].fd, &wfds)) n += 2;
				if(FD_ISSET(watches[i].fd, &efds)) n += 4;

				if(n) watches[i].callback(watches[i].fd, n, watches[i].data);
			}
		} else {
			XFlush(dsp);
		}
		display_timer_run();
	}
}

