#ifndef emu_internal_h
#define emu_internal_h

#include BUFFER
#include IO
#include DISPLAY

typedef enum {
	ES_NORMAL,
	ES_ESC,
	ES_SQUARE,
	ES_GETPARAMS,
	ES_FUNCKEY,
	ES_HASH,
	ES_SET_G0,
	ES_SET_G1,
	ES_PERCENT,
	ES_IGNORE,
	ES_NONSTD,
	ES_PALETTE,
} emu_state_t;

typedef enum {
	NUL = 0x00,
	BEL = 0x07,
	BS  = 0x08,
	HT  = 0x09,
	LF  = 0x0a,
	VT  = 0x0b,
	FF  = 0x0c,
	CR  = 0x0d,
	SO  = 0x0e,
	SI  = 0x0f,
	CAN = 0x18,
	SUB = 0x1a,
	ESC = 0x1b,
	DEL = 0x7f,
	CSI = 0x9b,
} emu_control_t;

#define NPARAM 16

typedef struct {
	/* io */
	void *io;
	
	/* buf */
	/* TODO: how do we know buf_state_t? */
	void *buf;
	buf_state_t *bs; /* cursor, attrs, size */
	
	/* esc sequence decode */
	emu_state_t es;
	unsigned int nparam;
	unsigned int param[NPARAM];
	unsigned int question; /* got ? in ESC sequence */
	unsigned int need_wrap; /* wrap needed at the end of the line */
	
	/* saved position */
	int saved_x;
	int saved_y;
	
	/* scrolling region */
	unsigned int top;
	unsigned int bottom;
	
	/* colors */
	buf_color_t saved_color; /* Saved foreground & background */
	
	/* modes */
	unsigned int crlf : 1;      /* CR+LF is needed */
	/* not implemented modes */
	unsigned int insert : 1;    /* insert */
	unsigned int charset : 1;   /* G0 / G1 */
	unsigned int saved_charset : 1; /* G0 / G1 */
	unsigned int show_control : 1;  /* display < 32 */
	unsigned int toggle_meta : 1;   /* toggle high bit */
	unsigned int decckm : 1;    /* cursor sends an extra 0 */
	unsigned int deccolm : 1;   /* 80/132 column */
	unsigned int decscnm : 1;   /* reverse video */
	unsigned int decom : 1;     /* origin is the corner of scroll region */
	unsigned int decawm : 1;    /* autowrap (default on) */
	unsigned int decarm : 1;    /* autorepeat (default on) */
	unsigned int mouse_x10 : 1; /* x10 mouse reporting */
	unsigned int mouse_x11 : 1; /* x11 mouse reporting */
	unsigned int deccm : 1;     /* cursor visible (default on) */
	
	unsigned int tab_stop[8];
} emu_t;

#ifdef NDEBUG
	#define D(s)
	#define NI(s)
#else
	#include <stdio.h>
	#define D(s) fprintf(stderr, "%s: %s\n", __func__, s)
	#define NI(s) fprintf(stderr, "%s: Not implemented (%s)\n", __func__, s)
#endif

/* linux.c */

void emu_setup_fd(int fd);
void *emu_create(void *buf);
void emu_destroy(void *e);
static void emu_reset_terminal(emu_t *e);

/* transmit.c */

static void emu_write_char(emu_t *e, emu_char_t c);
static void emu_respond_string(emu_t *e, const char *p);
void emu_dsp_char(void *emu, emu_char_t c);
void emu_io_char(void *emu, emu_char_t c);
void emu_dsp_ctrl(void *emu, ctrl_t ctrl);

/* process.c */

static inline void emu_csi_J(emu_t *e, int param);
static inline void emu_csi_K(emu_t *e, int param);
static inline void emu_csi_m(emu_t *e);
static inline void emu_set_mode(emu_t *e, int on_off);
static inline void emu_setterm_command(emu_t *e);
static inline void emu_process_params(emu_t *e, emu_char_t c);
static inline void emu_process(emu_t *e, emu_char_t c);

/* actions.c */

static inline void emu_clear_screen(emu_t *e);
static inline void emu_clear_lines(emu_t *e, int top, int bottom);
static inline void emu_clear_chars(emu_t *e, int start, int stop);
static inline void emu_scrollup(emu_t *e, int top, int bottom, unsigned int n);
static inline void emu_scrolldown(emu_t *e, int top, int bottom, unsigned int n);
static inline void emu_insert_chars(emu_t *e, unsigned int n);
static inline void emu_delete_chars(emu_t *e, unsigned int n);
static void emu_gotoxy(emu_t *e, int x, int y);
static void emu_gotoxay(emu_t *e, int x, int y);
static inline void emu_lf(emu_t *e);
static inline void emu_ri(emu_t *e);
static inline void emu_cr(emu_t *e);
static inline void emu_bs(emu_t *e);
static inline void emu_del(emu_t *e);
static inline void emu_cursor_report(emu_t *e);
static inline void emu_status_report(emu_t *e);
static inline void emu_respond_id(emu_t *e);
static void emu_save_cur(emu_t *e);
static void emu_restore_cur(emu_t *e);


#endif
